--
-- riot/Main.hs
-- 
-- Copyright (c) Tuomo Valkonen 2004.
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--

module Main(main) where

import IO(readFile, writeFile)
import Monad(liftM)
import System.Posix.User(getUserEntryForID, getRealUserID, homeDirectory)
import System.Posix.Signals as PosixSig
import Control.Exception(catchJust, ioErrors, throw, bracket)
import System.IO.Error(isDoesNotExistError)
import System(getArgs, getEnv)
import System.Console.GetOpt
import System.Exit
import Locale(setupLocale)

import qualified UI
import qualified Config
import Entry(EntryTree)
import Version(package, version)
import MBox

default_file=".riot"

get_home :: IO (String)
get_home =
    catch (getRealUserID >>= getUserEntryForID >>= (return . homeDirectory))
          (\_ -> getEnv "HOME")

get_default_file = do
    home <- get_home
    return $ home++"/"++default_file

do_load :: String -> IO [EntryTree MBoxEntry]
do_load fname = 
     liftM read_mbox_entrytree $ readFile fname

do_save :: String -> [EntryTree MBoxEntry] -> IO ()
do_save fname et = 
    writeFile fname $! show_mbox_entrytree et

load_initial :: String -> IO ([EntryTree MBoxEntry], String)
load_initial fname = do
    et <- catchJust ioErrors (do_load fname) exh
    return (et, fname)
    where
        exh e = if isDoesNotExistError e then
                    return []
                else
                    ioError e

load_default :: IO ([EntryTree MBoxEntry], String)
load_default =  get_default_file >>= load_initial


initui :: [EntryTree MBoxEntry] -> String -> IO (UI.Status MBoxEntry)
initui tt fname = do
    s <- UI.init Config.styles
    s <- return $ UI.set_topinfo s Config.topinfo_text
    s <- return $ UI.set_entries s tt (Just fname)
    s <- return $ UI.set_callbacks s do_save new_mboxentry
    return s

init_sighandlers s = do 
    PosixSig.setStoppedChildFlag True
    PosixSig.installHandler PosixSig.sigCHLD PosixSig.Default Nothing
    PosixSig.installHandler PosixSig.sigINT PosixSig.Ignore Nothing
    PosixSig.installHandler PosixSig.sigPIPE PosixSig.Ignore Nothing

release_sighandlers = do 
    PosixSig.installHandler PosixSig.sigINT PosixSig.Default Nothing
    PosixSig.installHandler PosixSig.sigPIPE PosixSig.Default Nothing


data Opts = Help | Version

options :: [OptDescr Opts]
options = [
    Option ['V']  ["version"] (NoArg Version) "Show version information",
    Option ['h']  ["help"]    (NoArg Help)    "Show this help"
    ]

usage = putStr $ usageInfo "Usage: riot [option...] [file]" options
versinfo = putStr $ package++" "++version++"\n"

do_opts :: [Opts] -> IO ()
do_opts [] = return ()
do_opts (o:oo) =
    case o of
        Help -> usage >> exitWith(ExitSuccess)
        Version -> versinfo >> exitWith(ExitSuccess)

do_args :: [String] -> IO ([EntryTree MBoxEntry], String)
do_args args =
    case (getOpt Permute options args) of
        (o, n, []) -> do
            do_opts o
            case n of
                [] -> load_default
                (f:[]) -> load_initial f
                otherwise -> error "Too many file parameters."
        (_, _, errs) -> error (concat errs)
        
main = do
    setupLocale
    args <- getArgs
    (tt, fname) <- do_args args
    tt <- return $! tt -- Force load before initialising UI
    bracket (initui tt fname >>= \s -> init_sighandlers s >> return s)
            (\_ -> UI.deinit >> release_sighandlers)
            (\s -> UI.refresh s >> UI.event_loop s Config.handle_key)

